//
//               INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//        Copyright(c) 2015 Intel Corporation. All Rights Reserved.

#include "tutorial_screen_capture.h"
#include "clock.h"

#include <thread>
#include <list>
#include <queue>
#include <mutex>

#define MSDK_ALIGN32(value)                      (((value + 31) >> 5) << 5) // round up to a multiple of 32
#define MSDK_ALIGN16(value)                      (((value + 15) >> 4) << 4) // round up to a multiple of 16

mfxStatus LoadFrame(mfxFrameSurface1& frame, FILE*& file, mfxFrameAllocator& mfxAllocator)
{
    if(!(frame.Data.Y && frame.Data.UV) && !frame.Data.MemId && !file)
        return MFX_ERR_NULL_PTR;

    bool unlock = false;
    if(frame.Data.MemId)
    {
        mfxAllocator.Lock(mfxAllocator.pthis, frame.Data.MemId, &frame.Data);
        unlock = true;
    }

    if(!((MFX_FOURCC_NV12 == frame.Info.FourCC  && MFX_CHROMAFORMAT_YUV420 == frame.Info.ChromaFormat) ||
         (MFX_FOURCC_RGB4 == frame.Info.FourCC  && MFX_CHROMAFORMAT_YUV444 == frame.Info.ChromaFormat) ||
         (DXGI_FORMAT_AYUV == frame.Info.FourCC && MFX_CHROMAFORMAT_YUV444 == frame.Info.ChromaFormat)))
        return MFX_ERR_UNSUPPORTED;


    mfxU32 w, h, i, pitch;
    mfxU32 nBytesRead;
    mfxU8 *ptr;

    if (frame.Info.CropH > 0 && frame.Info.CropW > 0) 
    {
        w = frame.Info.CropW;
        h = frame.Info.CropH;
    } 
    else 
    {
        w = frame.Info.Width;
        h = frame.Info.Height;
    }

    pitch = frame.Data.Pitch;

    ptr = frame.Data.Y + frame.Info.CropX + frame.Info.CropY * pitch;

    if(MFX_FOURCC_NV12 == frame.Info.FourCC)
    {
        for(i = 0; i < h; i++) 
        {
            nBytesRead = fread(ptr + i * pitch, 1, w, file);
            if(nBytesRead != w)
                return MFX_ERR_UNDEFINED_BEHAVIOR;
        }

        h     >>= 1;
        ptr = frame.Data.UV + frame.Info.CropX + (frame.Info.CropY >> 1) * pitch;
        for (i = 0; i < h; i++) 
        {
            nBytesRead = fread(ptr + i * pitch, 1, w, file);
            if(nBytesRead != w)
                return MFX_ERR_UNDEFINED_BEHAVIOR;
        }
    }
    else if (MFX_FOURCC_RGB4 == frame.Info.FourCC)
    {
        for (i = 0; i < h; i++) 
        {
            nBytesRead = fread(ptr + i * pitch, 1, 4*w, file);
            if(nBytesRead != 4*w)
                return MFX_ERR_UNDEFINED_BEHAVIOR;
        }
    }

    if(unlock)
        mfxAllocator.Unlock(mfxAllocator.pthis, frame.Data.MemId, &frame.Data);

    return MFX_ERR_NONE;
}

mfxStatus WriteFrame(mfxFrameSurface1& frame, FILE*& file, mfxFrameAllocator& mfxAllocator)
{
    if(!(frame.Data.Y && frame.Data.UV) && !frame.Data.MemId && !file)
        return MFX_ERR_NULL_PTR;

    bool unlock = false;
    if(frame.Data.MemId)
    {
        mfxAllocator.Lock(mfxAllocator.pthis, frame.Data.MemId, &frame.Data);
        unlock = true;
    }

    if(!((MFX_FOURCC_NV12 == frame.Info.FourCC  && MFX_CHROMAFORMAT_YUV420 == frame.Info.ChromaFormat) ||
         (MFX_FOURCC_RGB4 == frame.Info.FourCC  && MFX_CHROMAFORMAT_YUV444 == frame.Info.ChromaFormat) ||
         (DXGI_FORMAT_AYUV == frame.Info.FourCC && MFX_CHROMAFORMAT_YUV444 == frame.Info.ChromaFormat)))
        return MFX_ERR_UNSUPPORTED;


    mfxU32 w, h, i, pitch;
    mfxU32 nBytesRead;
    mfxU8 *ptr;

    if (frame.Info.CropH > 0 && frame.Info.CropW > 0) 
    {
        w = frame.Info.CropW;
        h = frame.Info.CropH;
    } 
    else 
    {
        w = frame.Info.Width;
        h = frame.Info.Height;
    }

    pitch = frame.Data.Pitch;

    ptr = frame.Data.Y + frame.Info.CropX + frame.Info.CropY * pitch;

    if(MFX_FOURCC_NV12 == frame.Info.FourCC)
    {
        for(i = 0; i < h; i++) 
        {
            nBytesRead = fwrite(ptr + i * pitch, 1, w, file);
            if(nBytesRead != w)
                return MFX_ERR_UNDEFINED_BEHAVIOR;
        }

        h     >>= 1;
        ptr = frame.Data.UV + frame.Info.CropX + (frame.Info.CropY >> 1) * pitch;
        for (i = 0; i < h; i++) 
        {
            nBytesRead = fwrite(ptr + i * pitch, 1, w, file);
            if(nBytesRead != w)
                return MFX_ERR_UNDEFINED_BEHAVIOR;
        }
    }
    else if ((MFX_FOURCC_RGB4 == frame.Info.FourCC) || (DXGI_FORMAT_AYUV == frame.Info.FourCC) )
    {
        ptr = min( min(frame.Data.R, frame.Data.G), frame.Data.B );
        ptr = ptr + frame.Info.CropX + frame.Info.CropY * pitch;

        for (i = 0; i < h; i++) 
        {
            nBytesRead = fwrite(ptr + i * pitch, 1, 4*w, file);
            if(nBytesRead != 4*w)
                return MFX_ERR_UNDEFINED_BEHAVIOR;
        }
    }

    if(unlock)
        mfxAllocator.Unlock(mfxAllocator.pthis, frame.Data.MemId, &frame.Data);

    return MFX_ERR_NONE;
}

mfxStatus AllocSysSurfaces(mfxFrameSurface1*& surf_pool, mfxFrameAllocRequest& request, mfxFrameAllocResponse& response)
{
    try
    {
        surf_pool = new mfxFrameSurface1[request.NumFrameSuggested];

        for(mfxU32 i = 0; i < (mfxU32) request.NumFrameSuggested; ++i)
        {
            memset(&surf_pool[i], 0, sizeof(surf_pool[i]));
            surf_pool[i].Info = request.Info;
            if(MFX_FOURCC_NV12 == surf_pool->Info.FourCC)
            {
                surf_pool[i].Data.Y = new mfxU8[surf_pool->Info.Width * request.Info.Height * 3 / 2];
                surf_pool[i].Data.UV = surf_pool[i].Data.Y + request.Info.Width * request.Info.Height;
                surf_pool[i].Data.Pitch = request.Info.Width;
            }
            else if(MFX_FOURCC_RGB4 == request.Info.FourCC || DXGI_FORMAT_AYUV == request.Info.FourCC)
            {
                surf_pool[i].Data.B = new mfxU8[request.Info.Width * request.Info.Height * 4];
                surf_pool[i].Data.G = surf_pool[i].Data.B + 1;
                surf_pool[i].Data.R = surf_pool[i].Data.B + 2;
                surf_pool[i].Data.A = surf_pool[i].Data.B + 3;
                surf_pool[i].Data.Pitch = 4 * request.Info.Width;
            }
        }
        memset(&response, 0, sizeof(response));
        response.NumFrameActual = request.NumFrameSuggested;

        return MFX_ERR_NONE;
    }
    catch(...)
    {
        return MFX_ERR_MEMORY_ALLOC;
    }
}

mfxStatus FreeSysSurfaces(mfxFrameSurface1*& surf_pool, mfxFrameAllocResponse& response)
{
    try
    {
        if(surf_pool)
        {
            for(mfxU32 i = 0; i < (mfxU32) response.NumFrameActual; ++i)
            {
                if(surf_pool[i].Data.Y)
                {
                    delete[] surf_pool[i].Data.Y;
                    surf_pool[i].Data.Y = 0;
                }
            }
            delete[] surf_pool;
            surf_pool = 0;
        }

        return MFX_ERR_NONE;
    }
    catch(...)
    {
        return MFX_ERR_MEMORY_ALLOC;
    }
}

void PrintHelp(char *strAppName, const char *strErrorMessage)
{
    printf("Intel(R) Media SDK Screen Capture Simple Sample\n\n");

    if (strErrorMessage)
    {
        printf("Error: %s\n", strErrorMessage);
    }

    printf("Usage: %s [Options]\n", strAppName);
    printf("Options: \n");
    printf("   [-n uint]           - number of frames to capture\n");
    printf("   [-w uint]           - screen resolution width\n");
    printf("   [-h uint]           - screen resolution height\n");
    printf("   [-new_w uint]       - resized width  (turns VPP on)\n");
    printf("   [-new_h uint]       - resized height (turns VPP on)\n");
    printf("   [-impl char]        - define library implementation ('d3d9' or 'd3d11')\n");
    printf("   [-fourcc char]      - fourcc format for captured from screen frame ('nv12' or 'rgb4')\n");
    printf("   [-o char]           - raw output filename (dump raw captured frame into this file)\n");
    printf("   [-enc]              - encode captured frames\n");
    printf("   [-enc_file char]    - specify encode output filename\n");
    printf("   [-enc_codec char]   - specify encode codec ('avc' or 'jpeg')\n");
    printf("   [-async_depth uint] - specify Async Depth\n");
    printf("   [-sys_mem]          - use system memory instead of video\n");
    printf("   [-?]                - print this help message\n");
    printf("\n");
}

mfxStatus ParseInputString(char* strInput[], mfxU8 nArgNum, sInputParams* pParams)
{
    if (1 == nArgNum)
    {
        PrintHelp(strInput[0], NULL);
        return MFX_ERR_UNSUPPORTED;
    }

    if(!pParams)
        return MFX_ERR_NULL_PTR;

    memset(pParams,0,sizeof(*pParams));

    for (mfxU32 i = 1; i < nArgNum; i++)
    {
        if (0 == strcmp(strInput[i], "-n"))
        {
            sscanf_s(strInput[++i], "%d", &pParams->n_frames);
        }
        else if (0 == strcmp(strInput[i], "-w"))
        {
            sscanf_s(strInput[++i], "%d", &pParams->width);
        }
        else if (0 == strcmp(strInput[i], "-h"))
        {
            sscanf_s(strInput[++i], "%d", &pParams->height);
        }
        else if (0 == strcmp(strInput[i], "-new_w"))
        {
            pParams->bVPP = true;
            sscanf_s(strInput[++i], "%d", &pParams->new_width);
        }
        else if (0 == strcmp(strInput[i], "-new_h"))
        {
            pParams->bVPP = true;
            sscanf_s(strInput[++i], "%d", &pParams->new_height);
        }
        else if (0 == strcmp(strInput[i], "-impl"))
        {
            if(i < nArgNum)
                ++i;
            else
                return MFX_ERR_UNDEFINED_BEHAVIOR;
            if(0 == strcmp(strInput[i], "d3d11"))
                pParams->impl = MFX_IMPL_VIA_D3D11;
            else if(0 == strcmp(strInput[i], "d3d9"))
                pParams->impl = MFX_IMPL_VIA_D3D9;
            else
            {
                PrintHelp(strInput[0], NULL);
                return MFX_ERR_UNSUPPORTED;
            }
        }
        else if (0 == strcmp(strInput[i], "-fourcc"))
        {
            if(i < nArgNum)
                ++i;
            else
                return MFX_ERR_UNDEFINED_BEHAVIOR;

            if(0 == strcmp(strInput[i], "nv12"))
                pParams->fourcc = MFX_FOURCC_NV12;
            if(0 == strcmp(strInput[i], "rgb4"))
                pParams->fourcc = MFX_FOURCC_RGB4;
        }
        else if (0 == strcmp(strInput[i], "-o"))
        {
            pParams->raw_output = true;
            strcpy(pParams->raw_output_file_name, strInput[++i]);
        }
        else if (0 == strcmp(strInput[i], "-enc"))
        {
            pParams->enc_output = true;
        }
        else if (0 == strcmp(strInput[i], "-enc_file"))
        {
            pParams->enc_output = true;
            strcpy(pParams->enc_output_file_name, strInput[++i]);
        }
        else if (0 == strcmp(strInput[i], "-delay_sync"))
        {
            pParams->delay_sync = true;
        }
        else if (0 == strcmp(strInput[i], "-sys_mem"))
        {
            pParams->system_memory = true;
        }
        else if (0 == strcmp(strInput[i], "-wait"))
        {
            sscanf_s(strInput[++i], "%d", &pParams->wait);
        }
        else if (0 == strcmp(strInput[i], "-limit_fps"))
        {
            sscanf_s(strInput[++i], "%d", &pParams->limit_fps);
        }
        else if (0 == strcmp(strInput[i], "-debug"))
        {
            pParams->debug = true;
        }
        else if (0 == strcmp(strInput[i], "-async_depth"))
        {
            sscanf_s(strInput[++i], "%d", &pParams->async_depth);
        }
        else if (0 == strcmp(strInput[i], "-enc_codec"))
        {
            if(i < nArgNum)
                ++i;
            else
                return MFX_ERR_UNDEFINED_BEHAVIOR;

            if((0 == strcmp(strInput[i], "AVC")) || (0 == strcmp(strInput[i], "avc")) || (0 == strcmp(strInput[i], "h264")))
                pParams->enc_codec = MFX_CODEC_AVC;
            if((0 == strcmp(strInput[i], "JPEG")) || (0 == strcmp(strInput[i], "jpeg")))
                pParams->enc_codec = MFX_CODEC_JPEG;
        }
        else if ((0 == strcmp(strInput[i], "-?")) || 0 == strcmp(strInput[i], "-h"))
        {
            PrintHelp(strInput[0], NULL);
            return MFX_ERR_UNSUPPORTED;
        }
    }

    //fill default values
    if(!pParams->n_frames)
        pParams->n_frames = 100;
    if(!pParams->fourcc)
        pParams->fourcc = MFX_FOURCC_NV12;
    if(!pParams->enc_codec)
        pParams->enc_codec = MFX_CODEC_AVC;
    if(!pParams->chroma_format)
    {
        if(MFX_FOURCC_NV12 == pParams->fourcc)
            pParams->chroma_format = MFX_CHROMAFORMAT_YUV420;
        else
            pParams->chroma_format = MFX_CHROMAFORMAT_YUV444;
    }
    if(!pParams->width)
        pParams->width = 1920;
    if(!pParams->height)
        pParams->height = 1080;
    if(!pParams->impl)
        pParams->impl = MFX_IMPL_VIA_D3D9;

    if(!pParams->async_depth)
        pParams->async_depth = 1;

    if(!pParams->new_width)
        pParams->new_width = pParams->width;
    if(!pParams->new_height)
        pParams->new_height = pParams->height;

    return MFX_ERR_NONE;
}

class FrameWriter
{
protected:
    bool end;
    std::queue<mfxFrameSurface1*> list;
    static std::mutex WriterMutex;

    mfxFrameAllocator& mfxAllocator;
    FILE*& file;

public:
    FrameWriter::FrameWriter(FILE*& _file, mfxFrameAllocator& _mfxAllocator)
        :file(_file),
        mfxAllocator(_mfxAllocator)
    {
        end = false;
    }
    FrameWriter::~FrameWriter()
    {
        end = true;
    }

    int PerformWrite();

    static int run(FrameWriter* that)
    {
        while(!that->end || that->list.size())
        {
            mfxFrameSurface1* toWrite = 0;
            Sleep(3);

            if(that->list.size())
            {
                std::lock_guard<std::mutex> lock(WriterMutex);
                toWrite = that->list.front();
                that->list.pop();
            }

            if(toWrite)
            {
                WriteFrame(*toWrite, that->file, that->mfxAllocator);
                toWrite->Data.Locked -= 1;
            }

        }

        return 0;

    }

    static int addFrameToQueue(FrameWriter* that, mfxFrameSurface1* pSurf)
    {
        std::lock_guard<std::mutex> lock(WriterMutex);

        that->list.push(pSurf);

        //pSurf->Data.Locked += 1;
        return 0;
    }
    //int removeFrameToQueue( mfxFrameSurface1* pSurf)
    //{
    //    std::lock_guard<std::mutex> lock(WriterMutex);

    //    list.push(pSurf);

    //    //pSurf->Data.Locked += 1;
    //}

    static int terminate(FrameWriter* that)
    {
        std::lock_guard<std::mutex> lock(WriterMutex);

        that->end = true;
        return 0;
    }

private:
    //prohobit copy constructor
    FrameWriter(const FrameWriter& that);
    //prohibit assignment operator
    FrameWriter& operator=(const FrameWriter&);

};


std::mutex FrameWriter::WriterMutex;

int main(int argc, char *argv[])
{
    std::cout << "Started\n";

    sInputParams params = {};
    mfxStatus mfxSts = MFX_ERR_NONE;

    mfxSts = ParseInputString(argv, argc, &params);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to parse input string");

    #ifdef _DEBUG
    if ( params.debug )
    {
        while ( !::IsDebuggerPresent() )
            ::Sleep( 100 );

        std::cout << "Debugger attached!!!\n";
        //::DebugBreak();
    }
    #endif

    if(strlen(params.raw_output_file_name))
    {
        params.raw_output = true;
    }

    if((params.new_width  && (params.width  != params.new_width)) ||
       (params.new_height && (params.height != params.new_height)) ||
        ((MFX_FOURCC_NV12 != params.fourcc) && (params.enc_output) && (MFX_CODEC_AVC == params.enc_codec)))
    {
        //VPP is required in case of color conversion and/or resize
        params.bVPP          = true;
    }

    mfxSession mfxSes = 0;

    mfxIMPL impl = MFX_IMPL_HARDWARE_ANY | params.impl;
    mfxVersion version;
    version.Major = MFX_VERSION_MAJOR;
    version.Minor = MFX_VERSION_MINOR;

    Clock initClock;
    initClock.Begin();

    mfxSts = MFXInit(impl,  &version, &mfxSes);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Init mfxSession");

    mfxHDL device_hdl = 0;
    mfxHandleType hdl_type = MFX_HANDLE_D3D9_DEVICE_MANAGER;

    //Init device manager
    if(MFX_IMPL_VIA_D3D9 == params.impl)
    {
        mfxSts = CreateDx9Device(mfxSes, &device_hdl, 0);
    }
    else if (MFX_IMPL_VIA_D3D11 == params.impl)
    {
        mfxSts = CreateDx11Device(mfxSes, &device_hdl, 0);
        hdl_type = MFX_HANDLE_D3D11_DEVICE;
    }
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Create device");

    //SetHandle
    mfxSts = MFXVideoCORE_SetHandle(mfxSes, hdl_type, device_hdl);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to SetHanlde");

    //prepare mfx allocator structure
    mfxFrameAllocator mfxAllocator;
    memset(&mfxAllocator, 0, sizeof(mfxAllocator));
    mfxAllocator.pthis  = &mfxAllocator; //Pointer to the allocator object, by mediasdk manual

    if(!params.system_memory)
    {
        if(MFX_IMPL_VIA_D3D9 == params.impl)
        {
            mfxAllocator.Alloc  = dx9_simple_alloc;
            mfxAllocator.Free   = dx9_simple_free;
            mfxAllocator.Lock   = dx9_simple_lock;
            mfxAllocator.Unlock = dx9_simple_unlock;
            mfxAllocator.GetHDL = dx9_simple_gethdl;
        } 
        else
        {
            mfxAllocator.Alloc  = dx11_simple_alloc;
            mfxAllocator.Free   = dx11_simple_free;
            mfxAllocator.Lock   = dx11_simple_lock;
            mfxAllocator.Unlock = dx11_simple_unlock;
            mfxAllocator.GetHDL = dx11_simple_gethdl;
        }

        // When using video memory we must provide Media SDK with an external allocator
        mfxSts = MFXVideoCORE_SetFrameAllocator(mfxSes, &mfxAllocator);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to SetFrameAllocator");
    }

    //Load plug-in
    mfxSts = MFXVideoUSER_Load(mfxSes, &MFX_PLUGINID_CAPTURE_HW, 1);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to load screen capture plug-in");

    //parameters for screen capture decoder init
    mfxVideoParam par;
    memset(&par,0,sizeof(par));
    par.AsyncDepth = params.async_depth;
    if(!params.system_memory)
    {
        par.IOPattern = MFX_IOPATTERN_OUT_VIDEO_MEMORY;
    }
    else
    {
        par.IOPattern = MFX_IOPATTERN_OUT_SYSTEM_MEMORY;
    }
    par.mfx.FrameInfo.Width  = MSDK_ALIGN16(params.width);
    par.mfx.FrameInfo.Height = MSDK_ALIGN16(params.height);
    par.mfx.FrameInfo.CropW = params.width;
    par.mfx.FrameInfo.CropH = params.height;
    par.mfx.FrameInfo.PicStruct = MFX_PICSTRUCT_PROGRESSIVE;
    par.mfx.FrameInfo.FrameRateExtN = 30;
    par.mfx.FrameInfo.FrameRateExtD = 1;
    par.mfx.FrameInfo.FourCC = params.fourcc;
    par.mfx.FrameInfo.ChromaFormat = params.chroma_format;
    par.mfx.CodecId = MFX_CODEC_CAPTURE;

    //parameters for encoder init
    mfxVideoParam enc_par;
    memset(&enc_par,0,sizeof(enc_par));
    enc_par.AsyncDepth = params.async_depth;
    if(!params.system_memory)
    {
        enc_par.IOPattern = MFX_IOPATTERN_IN_VIDEO_MEMORY;
    }
    else
    {
        enc_par.IOPattern = MFX_IOPATTERN_IN_SYSTEM_MEMORY;
    }
    enc_par.mfx.FrameInfo.PicStruct = MFX_PICSTRUCT_PROGRESSIVE;
    if(MFX_PICSTRUCT_PROGRESSIVE == enc_par.mfx.FrameInfo.PicStruct)
    {
        enc_par.mfx.FrameInfo.Width  = MSDK_ALIGN16(params.new_width);
        enc_par.mfx.FrameInfo.Height = MSDK_ALIGN16(params.new_height);
    }
    else
    {
        enc_par.mfx.FrameInfo.Width  = MSDK_ALIGN32(params.new_width);
        enc_par.mfx.FrameInfo.Height = MSDK_ALIGN32(params.new_height);
    }
    enc_par.mfx.FrameInfo.CropW = params.new_width;
    enc_par.mfx.FrameInfo.CropH = params.new_height;
    enc_par.mfx.FrameInfo.FrameRateExtN = 30;
    enc_par.mfx.FrameInfo.FrameRateExtD = 1;
    if(MFX_CODEC_JPEG == params.enc_codec && MFX_FOURCC_RGB4 == params.fourcc)
    {
        //JPEG encoder can encode from NV12, but for this tutorial, let's encode in JPEG from RGB4 without VPP
        enc_par.mfx.FrameInfo.FourCC = MFX_FOURCC_RGB4;
        enc_par.mfx.FrameInfo.ChromaFormat = MFX_CHROMAFORMAT_YUV444;
    } else {
        enc_par.mfx.FrameInfo.FourCC = MFX_FOURCC_NV12;
        enc_par.mfx.FrameInfo.ChromaFormat = MFX_CHROMAFORMAT_YUV420;
    }
    enc_par.mfx.CodecId = params.enc_codec;
    if(MFX_CODEC_AVC == enc_par.mfx.CodecId)
    {
        enc_par.mfx.RateControlMethod = MFX_RATECONTROL_CBR;
        enc_par.mfx.TargetUsage = MFX_TARGETUSAGE_BALANCED;
        enc_par.mfx.TargetKbps = 5000;
        enc_par.mfx.GopRefDist = 3;
    }
    else
    {
        enc_par.mfx.Quality= 75;
    }

    //parameters for VPP init
    mfxVideoParam vpp_par;
    memset(&vpp_par,0,sizeof(vpp_par));
    vpp_par.AsyncDepth = params.async_depth;
    vpp_par.IOPattern = par.IOPattern | enc_par.IOPattern;
    vpp_par.vpp.In = par.mfx.FrameInfo;
    vpp_par.vpp.Out = enc_par.mfx.FrameInfo;

    FILE* raw_output_file = 0;
    FILE* enc_output_file = 0;

    if(params.raw_output)
    {
        raw_output_file = fopen(params.raw_output_file_name, "wb");
        if(!raw_output_file)
            CHECK_STS(MFX_ERR_NOT_FOUND, MFX_ERR_NONE, "Failed to open file for raw output!");
    }

    if(params.enc_output && strlen(params.enc_output_file_name))
    {
        enc_output_file = fopen(params.enc_output_file_name, "wb");
        if(!enc_output_file)
            CHECK_STS(MFX_ERR_NOT_FOUND, MFX_ERR_NONE, "Failed to open file for encoded output!");
    }

    mfxSts = MFXVideoDECODE_Query(mfxSes, &par, &par);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Query DEcoder");

    mfxSts = MFXVideoDECODE_Init(mfxSes, &par);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Init DEcoder");

    if(params.enc_output)
    {
        mfxSts = MFXVideoENCODE_Query(mfxSes, &enc_par, &enc_par);
        if(MFX_WRN_PARTIAL_ACCELERATION == mfxSts)
            mfxSts = MFX_ERR_NONE;
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Query ENcoder");

        mfxSts = MFXVideoENCODE_Init(mfxSes, &enc_par);
        if(MFX_WRN_PARTIAL_ACCELERATION == mfxSts)
            mfxSts = MFX_ERR_NONE;
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Init ENcoder");
    }

    if(params.bVPP)
    {
        mfxSts = MFXVideoVPP_Query(mfxSes, &vpp_par, &vpp_par);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Query VPP");

        mfxSts = MFXVideoVPP_Init(mfxSes, &vpp_par);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Init VPP");
    }

    mfxFrameAllocRequest request;
    memset(&request, 0, sizeof(request));

    mfxFrameAllocRequest enc_request;
    memset(&enc_request, 0, sizeof(enc_request));

    mfxFrameAllocRequest vpp_request[2];
    memset(vpp_request, 0, 2*sizeof(mfxFrameAllocRequest));

    mfxFrameAllocRequest sum_request[2];
    memset(sum_request, 0, 2*sizeof(mfxFrameAllocRequest));

    mfxU32 frames_to_proceed = params.n_frames;
    mfxU32 frameNum = 0;

    mfxSts = MFXVideoDECODE_QueryIOSurf(mfxSes, &par, &request);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to QueryIOSurf decoder");

    request.NumFrameMin = request.NumFrameSuggested = 200 * request.NumFrameSuggested;

    if(params.enc_output)
    {
        mfxSts = MFXVideoENCODE_QueryIOSurf(mfxSes, &enc_par, &enc_request);
        if(MFX_WRN_PARTIAL_ACCELERATION == mfxSts)
            mfxSts = MFX_ERR_NONE;
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to QueryIOSurf encode");
    }

    if(params.bVPP)
    {
        mfxSts = MFXVideoVPP_QueryIOSurf(mfxSes, &vpp_par, vpp_request);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to QueryIOSurf vpp");
    }

    sum_request[0].Info              = request.Info;
    sum_request[0].Type              = request.Type              | vpp_request[0].Type;
    sum_request[0].NumFrameMin       = request.NumFrameMin       + vpp_request[0].NumFrameMin;
    sum_request[0].NumFrameSuggested = request.NumFrameSuggested + vpp_request[0].NumFrameSuggested;

    if(params.bVPP)
    {
        sum_request[1].Info              = vpp_request[1].Info;
        sum_request[1].Type              = enc_request.Type              | vpp_request[1].Type;
        sum_request[1].NumFrameMin       = enc_request.NumFrameMin       + vpp_request[1].NumFrameMin;
        sum_request[1].NumFrameSuggested = enc_request.NumFrameSuggested + vpp_request[1].NumFrameSuggested;
    }
    else if(params.enc_output)
    {
        sum_request[0].Type              |= enc_request.Type;
        sum_request[0].NumFrameMin       += enc_request.NumFrameMin;
        sum_request[0].NumFrameSuggested += enc_request.NumFrameSuggested;
    }

    if ((sum_request[0].Type & MFX_MEMTYPE_DXVA2_PROCESSOR_TARGET) && ((sum_request[0].Type & 0xf0) != MFX_MEMTYPE_DXVA2_PROCESSOR_TARGET))
        sum_request[0].Type = (mfxU16)(sum_request[0].Type & (~ MFX_MEMTYPE_DXVA2_PROCESSOR_TARGET));

    if ((sum_request[1].Type & MFX_MEMTYPE_DXVA2_PROCESSOR_TARGET) && ((sum_request[1].Type & 0xf0) != MFX_MEMTYPE_DXVA2_PROCESSOR_TARGET))
        sum_request[1].Type = (mfxU16)(sum_request[1].Type & (~ MFX_MEMTYPE_DXVA2_PROCESSOR_TARGET));

    //Allocate frames
    mfxFrameSurface1* dec_surf_pool = 0;
    mfxFrameSurface1* vpp_surf_pool = 0;

    mfxFrameAllocResponse responce[2];
    memset(responce, 0, 2*sizeof(mfxFrameAllocResponse));

    if(!params.system_memory)
    {
        mfxSts = mfxAllocator.Alloc(mfxAllocator.pthis, &sum_request[0], &responce[0]);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Alloc frames");
        dec_surf_pool = new mfxFrameSurface1[responce[0].NumFrameActual];
        for(mfxU32 i = 0; i < (mfxU32) responce[0].NumFrameActual; ++i)
        {
            memset(&dec_surf_pool[i], 0, sizeof(dec_surf_pool[i]));
            dec_surf_pool[i].Info = sum_request[0].Info;
            dec_surf_pool[i].Data.MemId = responce[0].mids[i];
        }
    }
    else
    {
        mfxSts = AllocSysSurfaces(dec_surf_pool, sum_request[0], responce[0]);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Alloc system frames");
    }

    if(params.bVPP)
    {
        if(!params.system_memory)
        {
            mfxSts = mfxAllocator.Alloc(mfxAllocator.pthis, &sum_request[1], &responce[1]);
            CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Alloc frames");
            vpp_surf_pool = new mfxFrameSurface1[responce[1].NumFrameActual];
            for(mfxU32 i = 0; i < (mfxU32) responce[1].NumFrameActual; ++i)
            {
                memset(&vpp_surf_pool[i], 0, sizeof(vpp_surf_pool[i]));
                vpp_surf_pool[i].Info = sum_request[1].Info;
                vpp_surf_pool[i].Data.MemId = responce[1].mids[i];
            }
        }
        else
        {
            mfxSts = AllocSysSurfaces(dec_surf_pool, sum_request[1], responce[1]);
            CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Alloc system frames");
        }
    }

    mfxBitstream bitstream;
    memset(&bitstream, 0, sizeof(bitstream));

    if(params.enc_output)
    {
        mfxSts = MFXVideoENCODE_GetVideoParam(mfxSes, &enc_par);
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to GetVideoParam from Encoder");
        if(!enc_par.mfx.BufferSizeInKB)
        {
            enc_par.mfx.BufferSizeInKB = 10240;
        }

        bitstream.MaxLength = enc_par.mfx.BufferSizeInKB*1000*max(1, enc_par.mfx.BRCParamMultiplier) * params.async_depth;
        bitstream.Data = new mfxU8[bitstream.MaxLength];
    }

    mfxSyncPoint dec_syncpoint = 0;
    mfxSyncPoint vpp_syncpoint = 0;
    mfxSyncPoint enc_syncpoint = 0;

    mfxU32 async = max(1, params.async_depth - 1);
    mfxU32 submitted = 0;

    initClock.End();
    Clock decClock;
    Clock vppClock;
    Clock encClock;

    std::cout << "Init succeed\n";

    std::cout << "Init time: " << initClock.Duration() << " ms \n";

    //std::thread* pWriter = 0;

    FrameWriter CWriter(raw_output_file, mfxAllocator);
    std::thread writer(FrameWriter::run, &CWriter);

    double limitFPSms = 1000.0 / params.limit_fps;
    double since_last = 0;
    Clock limitFPS;
    limitFPS.Begin();

    Clock totalClock;
    totalClock.Begin();
    //actual processing, syncronous pipeline
    while (frameNum < frames_to_proceed )
    {
        mfxFrameSurface1* p_frame = 0;
        mfxFrameSurface1* p_out_frame = 0;
        mfxFrameSurface1* p_vpp_frame = 0;
        //find free surface
        //Sleep(params.wait);

        for(mfxU32 j = 0; j < 1000; ++j)
        {
            for(mfxU32 i = 0; i < responce[0].NumFrameActual; ++i)
            {
                if(!dec_surf_pool[i].Data.Locked)
                {
                    p_frame = &dec_surf_pool[i];
                    break;
                }
            }
            if(p_frame)
                break;
            Sleep(1);
        }

        if(!p_frame)
        {
            CHECK_STS(MFX_ERR_UNDEFINED_BEHAVIOR, MFX_ERR_NONE, "Failed: all decoder's frames are locked!");
        }

        decClock.Begin();
        since_last = limitFPS.TimeSinceLastCall();
        if(since_last < limitFPSms)
            Sleep(limitFPSms - since_last - 1);
        mfxSts = MFXVideoDECODE_DecodeFrameAsync(mfxSes, 0, p_frame, &p_out_frame, &dec_syncpoint);
        limitFPS.TimeSinceLastCall();

        if(MFX_ERR_MORE_DATA == mfxSts)
        {
            continue;
        }
        if(MFX_ERR_MORE_SURFACE == mfxSts || mfxSts > 0)
        {
            continue;
        }
        CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to DecodeFrameAsync");

        if(!params.delay_sync)
        {
            mfxSts = MFXVideoCORE_SyncOperation(mfxSes, dec_syncpoint, MFX_INFINITE);
            CHECK_STS(mfxSts, MFX_ERR_NONE, "Decode SyncOperation failed");
            dec_syncpoint = 0;
        }

        if(raw_output_file)
        {
            if(params.delay_sync)
            {
                mfxSts = MFXVideoCORE_SyncOperation(mfxSes, dec_syncpoint, MFX_INFINITE);
                dec_syncpoint = 0;
                CHECK_STS(mfxSts, MFX_ERR_NONE, "Decode SyncOperation failed");
            }

            //if(pWriter)
            //{
            //    pWriter->join();
            //    delete pWriter;
            //}
            //pWriter = new std::thread(WriteFrame, *p_out_frame, raw_output_file, mfxAllocator);
            p_out_frame->Data.Locked += 1;
            CWriter.addFrameToQueue(&CWriter, p_out_frame);

            //mfxSts = WriteFrame(*p_out_frame, raw_output_file, mfxAllocator);
            CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to write next output raw frame");
        }
        decClock.End();

        if(params.bVPP)
        {
            //find free surface
            for(mfxU32 i = 0; i < responce[1].NumFrameActual; ++i)
            {
                if(!vpp_surf_pool[i].Data.Locked)
                {
                    p_vpp_frame = &vpp_surf_pool[i];
                    break;
                }
            }
            if(!p_vpp_frame)
            {
                CHECK_STS(MFX_ERR_UNDEFINED_BEHAVIOR, MFX_ERR_NONE, "Failed: all vpp's frames are locked!");
            }

            vppClock.Begin();
            mfxSts = MFXVideoVPP_RunFrameVPPAsync(mfxSes, p_out_frame, p_vpp_frame, 0, &vpp_syncpoint);
            if(MFX_ERR_MORE_DATA == mfxSts)
            {
                continue;
            }
            if(MFX_ERR_MORE_SURFACE == mfxSts || mfxSts > 0)
            {
                continue;
            }
            CHECK_STS(mfxSts, MFX_ERR_NONE, "VPP Operation failed");

            if(!params.delay_sync)
            {
                mfxSts = MFXVideoCORE_SyncOperation(mfxSes, vpp_syncpoint, MFX_INFINITE);
                CHECK_STS(mfxSts, MFX_ERR_NONE, "VPP SyncOperation failed");
                vpp_syncpoint = 0;
            }
            vppClock.End();
        }
        else
        {
            p_vpp_frame = p_out_frame;
        }

        if(params.enc_output)
        {
            encClock.Begin();
            mfxSts = MFXVideoENCODE_EncodeFrameAsync(mfxSes, 0, p_vpp_frame, &bitstream, &enc_syncpoint);
            if(MFX_ERR_MORE_DATA == mfxSts)
            {
                continue;
            }
            if(MFX_ERR_MORE_SURFACE == mfxSts || mfxSts > 0)
            {
                continue;
            }
            CHECK_STS(mfxSts, MFX_ERR_NONE, "EncodeFrameAsync failed");

            if(MFX_ERR_NONE == mfxSts)
            {
                submitted++;
            }

            if(submitted >= async)
            {
                mfxSts = MFXVideoCORE_SyncOperation(mfxSes, enc_syncpoint, MFX_INFINITE);
                CHECK_STS(mfxSts, MFX_ERR_NONE, "Encode SyncOperation failed");

                //dump bitstream
                if(enc_output_file)
                {
                    mfxU32 written = 0;
                    written = fwrite(bitstream.Data+bitstream.DataOffset, sizeof(mfxU8), bitstream.DataLength, enc_output_file);
                    if(written != bitstream.DataLength)
                    {
                        CHECK_STS(MFX_ERR_UNDEFINED_BEHAVIOR, MFX_ERR_NONE, "Failed to write encoded stream");
                    }
                }
                bitstream.DataOffset = 0;
                bitstream.DataLength = 0;

                //by manual
                /* The SDK simplifies the requirement for asynchronous pipeline synchronization. The application only needs to synchronize after the last SDK function. 
                    Explicit synchronization of intermediate results is not required and in fact can slow performance */
                //nulling syncpoints is not required, just to assert that no more sync is required
                dec_syncpoint = 0;
                enc_syncpoint = 0;
                vpp_syncpoint = 0;
            }
            encClock.End();
        }

        frameNum++;
    }
    totalClock.End();

    FrameWriter::terminate(&CWriter);
    writer.join();

    if(1 == params.async_depth)
    {
        std::cout << "\nDec time: " << decClock.Duration() << " ms\n";
        if(params.bVPP)
            std::cout << "VPP time: " << vppClock.Duration() << " ms\n";
        if(params.enc_output)
            std::cout << "Enc time: " << encClock.Duration() << " ms\n";
        if(params.bVPP || params.enc_output)
            std::cout << "Sum time: " << decClock.Duration() + vppClock.Duration() + encClock.Duration() << " ms\n";
    }
    std::cout << "\nTotal time: " << totalClock.Duration() << " ms\n";


    std::cout << "\nAverage time for 1 frame:\n";
    if(1 == params.async_depth)
    {
        std::cout << "Dec: " << (double) (decClock.Duration() / frameNum) << " ms\n";
        if(params.bVPP)
            std::cout << "VPP: " << (double) (vppClock.Duration() / frameNum) << " ms\n";
        if(params.enc_output)
            std::cout << "Enc: " << (double) (encClock.Duration() / frameNum) << " ms\n";
        if(params.bVPP || params.enc_output)
            std::cout << "Sum: " << (double) ( (decClock.Duration() + vppClock.Duration() + encClock.Duration()) / frameNum ) << " ms\n";
    }
    std::cout << "Total average: " << (double) (totalClock.Duration() / frameNum) << " ms\n";

    std::cout << "\nFPS:\n";
    if(1 == params.async_depth)
    {
        std::cout << "Dec: " << (double) (1000 * frameNum / decClock.Duration()) << " \n";
        if(params.bVPP)
            std::cout << "VPP: " << (double) (1000 * frameNum / vppClock.Duration()) << " \n";
        if(params.enc_output)
            std::cout << "Enc: " << (double) (1000 * frameNum / encClock.Duration()) << " \n";
        if(params.bVPP || params.enc_output)
            std::cout << "Sum: " << (double) (1000 * frameNum / (decClock.Duration() + vppClock.Duration() + encClock.Duration()) ) << " \n";
    }
    std::cout << "Total FPS: " << (double) (1000 * frameNum / totalClock.Duration()) << " \n";

    Clock closeClock;
    closeClock.Begin();

    //Free resources
    if(raw_output_file)
    {
        fclose(raw_output_file);
        raw_output_file = 0;
    }
    if(enc_output_file)
    {
        fclose(enc_output_file);
        enc_output_file = 0;
    }

    mfxSts = MFXVideoDECODE_Close(mfxSes);
    if(mfxSts)
        std::cout << "Decode close failed, returned status: " << mfxSts << "\n";

    if(params.bVPP)
    {
        mfxSts = MFXVideoVPP_Close(mfxSes);
        if(mfxSts)
            std::cout << "VPP close failed, returned status: " << mfxSts << "\n";
    }

    if(params.enc_output)
    {
        mfxSts = MFXVideoENCODE_Close(mfxSes);
        if(mfxSts)
            std::cout << "Encode close failed, returned status: " << mfxSts << "\n";

        if(bitstream.Data)
        {
            delete[] bitstream.Data;
            bitstream.Data = 0;
        }
    }

    if(!params.system_memory)
    {
        mfxSts = mfxAllocator.Free(mfxAllocator.pthis, &responce[0]);
    }
    else
    {
        mfxSts = FreeSysSurfaces(dec_surf_pool, responce[0]);
    }
    if(mfxSts)
        std::cout << "Failed to free decoder's surface pool: " << mfxSts << "\n";

    if(params.bVPP)
    {
        if(!params.system_memory)
        {
            mfxSts = mfxAllocator.Free(mfxAllocator.pthis, &responce[1]);
        }
        else
        {
            mfxSts = FreeSysSurfaces(vpp_surf_pool, responce[1]);
        }
        if(mfxSts)
            std::cout << "Failed to free VPP's surface pool: " << mfxSts << "\n";
    }

    if(dec_surf_pool)
    {
        delete[] dec_surf_pool;
        dec_surf_pool = 0;
    }

    if(vpp_surf_pool)
    {
        delete[] vpp_surf_pool;
        vpp_surf_pool = 0;
    }

    mfxSts = MFXVideoUSER_UnLoad(mfxSes, &MFX_PLUGINID_CAPTURE_HW);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Unload plugin");

    mfxSts = MFXClose(mfxSes);
    CHECK_STS(mfxSts, MFX_ERR_NONE, "Failed to Close mfxSession");
    mfxSes = 0;

    if(MFX_IMPL_VIA_D3D11 == params.impl)
    {
        CleanupDx11Device();
    }
    else if (MFX_IMPL_VIA_D3D9 == params.impl)
    {
        CleanupDx9Device();
    }

    closeClock.End();

    std::cout << "\nClose time: " << closeClock.Duration() << " ms \n\n";

    std::cout << "Processed " << frameNum << " frames\n";
    std::cout << "Done!\n";

    return 0;
}